"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const XRegExp = require("xregexp");
const gensequence_1 = require("gensequence");
const Text = require("./text");
const fileReader_1 = require("./fileReader");
const cspell_lib_1 = require("cspell-lib");
const rxjs_1 = require("rxjs");
const operators_1 = require("rxjs/operators");
const path = require("path");
const fs_extra_1 = require("fs-extra");
const Trie = require("cspell-trie");
const HR = require("hunspell-reader");
const iterateWordsFromFile_1 = require("./iterateWordsFromFile");
const fileWriter_1 = require("./fileWriter");
const regNonWordOrSpace = XRegExp("[^\\p{L}' ]+", 'gi');
const regExpSpaceOrDash = /(?:\s+)|(?:-+)/g;
const regExpRepeatChars = /(.)\1{3,}/i;
function normalizeWords(lines) {
    return lines.concatMap(line => lineToWords(line));
}
exports.normalizeWords = normalizeWords;
function lineToWords(line) {
    // Remove punctuation and non-letters.
    const filteredLine = line.replace(regNonWordOrSpace, '|');
    const wordGroups = filteredLine.split('|');
    const words = gensequence_1.genSequence(wordGroups)
        .concatMap(a => [a, ...a.split(regExpSpaceOrDash)])
        .concatMap(a => splitCamelCase(a))
        .map(a => a.trim())
        .filter(a => !!a)
        .filter(s => !regExpRepeatChars.test(s))
        .map(a => a.toLowerCase());
    return words;
}
exports.lineToWords = lineToWords;
function splitCamelCase(word) {
    const splitWords = Text.splitCamelCaseWord(word);
    // We only want to preserve this: "New York" and not "Namespace DNSLookup"
    if (splitWords.length > 1 && regExpSpaceOrDash.test(word)) {
        return gensequence_1.genSequence(splitWords).concatMap(w => w.split(regExpSpaceOrDash));
    }
    return splitWords;
}
async function compileWordList(filename, destFilename, options) {
    const getWords = () => regHunspellFile.test(filename) ? readHunspellFiles(filename) : fileReader_1.lineReaderRx(filename);
    const destDir = path.dirname(destFilename);
    const pDir = fs_extra_1.mkdirp(destDir);
    const stream = getWords().pipe(options.splitWords ? compileWordListWithSplitRx : compileSimpleWordListRx, operators_1.filter(a => !!a), operators_1.distinct(), options.sort ? sort : operators_1.map(a => a));
    const buffered = stream.pipe(operators_1.map(a => a + '\n'), operators_1.bufferCount(1024), operators_1.map(a => a.join('')));
    await pDir;
    return cspell_lib_1.writeToFileRxP(destFilename, buffered);
}
exports.compileWordList = compileWordList;
function compileWordListWithSplit(filename, destFilename) {
    return compileWordList(filename, destFilename, { splitWords: true, sort: true });
}
exports.compileWordListWithSplit = compileWordListWithSplit;
async function compileSimpleWordList(filename, destFilename, options) {
    return compileWordList(filename, destFilename, { splitWords: false, sort: true });
}
exports.compileSimpleWordList = compileSimpleWordList;
function sort(words) {
    return words.pipe(operators_1.toArray(), operators_1.flatMap(a => a.sort()));
}
function compileWordListWithSplitRx(words) {
    return words.pipe(operators_1.flatMap(line => lineToWords(line).toArray()));
}
function compileSimpleWordListRx(words) {
    return words.pipe(operators_1.map(a => a.toLowerCase()));
}
function normalizeWordsToTrie(words) {
    const result = normalizeWords(words)
        .reduce((node, word) => Trie.insert(word, node), {});
    return result;
}
exports.normalizeWordsToTrie = normalizeWordsToTrie;
async function compileWordListToTrieFile(words, destFilename) {
    const destDir = path.dirname(destFilename);
    const pDir = fs_extra_1.mkdirp(destDir);
    const pRoot = normalizeWordsToTrie(words);
    const [_, root] = await Promise.all([pDir, pRoot]);
    return fileWriter_1.writeSeqToFile(Trie.serializeTrie(root, { base: 32, comment: 'Built by cspell-tools.' }), destFilename);
}
exports.compileWordListToTrieFile = compileWordListToTrieFile;
const regHunspellFile = /\.(dic|aff)$/i;
function readHunspellFiles(filename) {
    const dicFile = filename.replace(regHunspellFile, '.dic');
    const affFile = filename.replace(regHunspellFile, '.aff');
    const reader = HR.HunspellReader.createFromFiles(affFile, dicFile);
    const r = rxjs_1.from(reader).pipe(operators_1.flatMap(reader => reader.readWordsRx()), operators_1.map(aff => aff.word));
    return r;
}
async function compileTrie(filename, destFilename) {
    const words = await iterateWordsFromFile_1.streamWordsFromFile(filename);
    return compileWordListToTrieFile(words, destFilename);
}
exports.compileTrie = compileTrie;
//# sourceMappingURL=wordListCompiler.js.map